/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.voxelisation.tls;

import fr.amap.lidar.format.jleica.LPointShotExtractor;
import fr.amap.lidar.format.jleica.LShot;
import fr.amap.lidar.format.jleica.ptx.PTXScan;
import fr.amap.commons.math.matrix.Mat3D;
import fr.amap.commons.math.matrix.Mat4D;
import fr.amap.commons.math.vector.Vec3D;
import fr.amap.commons.math.vector.Vec4D;
import fr.amap.commons.util.CallableTaskAdapter;
import fr.amap.commons.util.IteratorWithException;
import fr.amap.commons.util.ProcessingAdapter;
import fr.amap.lidar.amapvox.commons.LidarScan;
import fr.amap.lidar.amapvox.shot.Shot;
import fr.amap.lidar.amapvox.util.AVoxTask;
import fr.amap.lidar.amapvox.util.FinalCounter;
import fr.amap.lidar.amapvox.voxelisation.configuration.PTXLidarScan;
import fr.amap.lidar.amapvox.voxelisation.configuration.TLSVoxCfg;
import fr.amap.lidar.amapvox.voxelisation.configuration.VoxMergingCfg;
import fr.amap.lidar.amapvox.voxelisation.postproc.VoxelFileMerging;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;
import org.apache.log4j.Logger;

/**
 *
 * @author Julien Heurtebize
 */
public class PTXVoxelisation extends AVoxTask {

    private final static Logger LOGGER = Logger.getLogger(PTXVoxelisation.class);

    public PTXVoxelisation(File file, int ncpu) {
        super(file, ncpu);
    }

    @Override
    public String getName() {
        return "TLS voxelisation";
    }

    @Override
    public void init() throws Exception {
        // nothing to do
    }

    @Override
    public File[] call() throws Exception {

        TLSVoxCfg mainCfg = new TLSVoxCfg();
        mainCfg.readConfiguration(file);
        List<LidarScan> lidarScans = mainCfg.getLidarScans();
        ArrayList<File> files = new ArrayList<>();
        ExecutorService exec = Executors.newFixedThreadPool(ncpu);
        FinalCounter nDone = new FinalCounter(0);
        try {
            List<SingleScanPTXVoxelisation> tasks = new ArrayList();
            int count = 0;
            for (LidarScan scan : lidarScans) {

                TLSVoxCfg cfg = new TLSVoxCfg();
                cfg.readConfiguration(file);
                PTXLidarScan ptxScan = (PTXLidarScan) scan;
                cfg.setInputFile(scan.file);
                File outputFile = new File(mainCfg.getOutputFile().getAbsolutePath() + "/" + scan.file.getName() + "-scan-" + count + ".vox");
                cfg.setOutputFile(outputFile);

                File tmpFile = File.createTempFile(scan.file.getName(), ".xml");
                cfg.writeConfiguration(tmpFile, cfg.getBuildVersion());
                SingleScanPTXVoxelisation ptxVoxelization = new SingleScanPTXVoxelisation(ptxScan.getScan(), tmpFile);
                ptxVoxelization.init();
                ptxVoxelization.addCallableTaskListener(new CallableTaskAdapter() {
                    @Override
                    public void onSucceeded() {
                        nDone.increment();
                        String msg = "Voxelization, scan " + nDone.getValue() + "/" + lidarScans.size();
                        fireProgress(msg, nDone.getValue(), lidarScans.size());
                    }
                });

                tasks.add(ptxVoxelization);
                files.add(outputFile);
                count++;
            }

            // wait for every scan voxelisation to finish
            LOGGER.info("Voxelization...");
            List<Future<File[]>> results = exec.invokeAll(tasks);
            for (Future<File[]> result : results) {
                files.addAll(Arrays.asList(result.get()));
            }

            if (mainCfg.getVoxelParameters().isMergingAfter()) {

                LOGGER.info("Merging voxel files...");

                VoxMergingCfg mergingCfg = new VoxMergingCfg(mainCfg.getVoxelParameters().getMergedFile(), mainCfg.getVoxelParameters(), files);
                File mergingCfgFile = File.createTempFile(file.getName() + "-merging-", ".xml");
                mergingCfg.writeConfiguration(mergingCfgFile, mainCfg.getBuildVersion());
                VoxelFileMerging tool = new VoxelFileMerging(mergingCfgFile, 1);

                tool.addProcessingListener(new ProcessingAdapter() {
                    @Override
                    public void processingStepProgress(String progressMsg, long progress, long max) {
                        fireProgress(progressMsg, progress, max);
                    }
                });

                tool.init();
                files.addAll(Arrays.asList(tool.call()));
            }

        } catch (InterruptedException | NullPointerException ex) {
            fireCancelled();
            throw ex;
        }

        return files.toArray(new File[files.size()]);
    }

    class SingleScanPTXVoxelisation extends TLSVoxelisation {

        private final Logger LOGGER = Logger.getLogger(PTXVoxelisation.class);

        private final PTXScan scan;

        public SingleScanPTXVoxelisation(PTXScan scan, File file) throws Exception {
            super(file, 1);
            this.scan = scan;
        }

        @Override
        public String getName() {
            return "TLS voxelisation (PTX)";
        }

        @Override
        public File[] call() throws Exception {

            LOGGER.info("PTX extraction is started");

            Iterator<LShot> iterator = new LPointShotExtractor(scan).iterator();

            IteratorWithException<fr.amap.lidar.amapvox.shot.Shot> it = new IteratorWithException() {

                private int index = 0;

                @Override
                public boolean hasNext() throws Exception {
                    return iterator.hasNext();
                }

                @Override
                public Shot next() throws Exception {
                    LShot shot = iterator.next();
                    if (shot != null) {
                        Vec4D locVector = Mat4D.multiply(transfMatrix, new Vec4D(shot.origin.x, shot.origin.y, shot.origin.z, 1.0d));
                        Vec3D uVector = Mat3D.multiply(rotation, new Vec3D(shot.direction.x, shot.direction.y, shot.direction.z));
                        return new Shot(index++, new Point3d(locVector.x, locVector.y, locVector.z), new Vector3d(uVector.x, uVector.y, uVector.z), shot.ranges);
                    } else {
                        return null;
                    }
                }
            };

            File[] files = voxelAnalysis.voxelization(it);

            fireSucceeded();

            return files;
        }

    }
}
