/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.voxels;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Voxel space file parsing class.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class VoxelSpaceFile {
    
    private final List<String> names;
    private final Map<String, Float> max;
    private final Map<String, Float> min;
    private final List<List<String>> lines;
    private int width, height, depth;
    
    public VoxelSpaceFile() {
        this.names = new ArrayList();
        this.lines = new ArrayList();
        this.max = new HashMap();
        this.min = new HashMap();
        this.width = 0;
        this.height = 0;
        this.depth = 0;
    }
    
    public void setSize(int width, int height, int depth) {
        this.width = width;
        this.height = height;
        this.depth = depth;
    }
    
    public int getWidth() {
        return width;
    }
    
    public int getHeight() {
        return height;
    }
    
    public int getDepth() {
        return depth;
    }
    
    public final void addName(final String name) {
        names.add(name);
    }
    
    public final void addLine(final List<String> line) {
        lines.add(line);
    }
    
    public String[] getNames() {
        String[] array = new String[names.size()];
        
        for (int i = 0; i < array.length; i++) {
            array[i] = names.get(i);
        }
        
        return array;
    }
    
    public final String getProperty(String name, int index) {
        if (lines.size() > index) {
            if (names.contains(name)) {
                return lines.get(index).get(names.indexOf(name));
            }
        }
        
        return null;
    }
    
    public final float getMinPropertyValue(String name) {
        return min.get(name);
    }
    
    public final float getMaxPropertyValue(String name) {        
        return max.get(name);
    }
    
    public final List<String> getLine(String name) {
        List<String> values = new ArrayList<>();
        
        for (int i = 0; i < lines.size(); i++) {
            values.add(getProperty(name, i));
        }
        
        return values;
    }
    
    public static VoxelSpaceFile parse(final InputStream voxelFile) {
        final VoxelSpaceFile parsed = new VoxelSpaceFile();
        
        if (voxelFile != null) {
            try {
                final BufferedReader br = new BufferedReader(new InputStreamReader(voxelFile));
                
                if (br.ready()) {
                    if (br.readLine().contains("VOXEL SPACE")) {
                        br.lines().forEach((String line) -> {
                            if (!line.startsWith("#") && !line.startsWith("VOXEL SPACE")) {
                                if (line.startsWith("i j k")) {
                                    for (String name : line.split(" ")) {
                                        parsed.addName(name);
                                    }
                                } else {
                                    List<String> values = new ArrayList<>();
                                    
                                    for (String value : line.split(" ")) {
                                        values.add(value);
                                    }
                                    
                                    parsed.addLine(values);
                                }
                            } else if (line.startsWith("#split")) {
                                String[] values = line.substring(8).split(" ");
                                
                                parsed.setSize(Integer.parseInt(values[0]),
                                        Integer.parseInt(values[1]),
                                        Integer.parseInt(values[2]));
                            }
                        });
                        
                        parsed.names.forEach((name) -> {
                            for (int i = 0; i < parsed.lines.size(); i++) {
                                float value = Float.valueOf(parsed.getProperty(name, i));

                                if (!Float.isNaN(value)) {
                                    if (!parsed.max.containsKey(name)) {
                                        parsed.max.put(name, value);
                                    } else if (value > parsed.max.get(name)) {
                                        parsed.max.put(name, value);
                                    }
                                    
                                    if (!parsed.min.containsKey(name)) {
                                        parsed.min.put(name, value);
                                    } else if (value < parsed.min.get(name)) {
                                        parsed.min.put(name, value);
                                    }
                                }
                            }
                        });
                    } else {
                        System.err.println("The voxel space file doesn't "
                                + "have a 'Voxel Space' header.");
                        return null;
                    }
                }
            } catch (FileNotFoundException ex) {
                Logger.getLogger(VoxelSpaceFile.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IOException ex) {
                Logger.getLogger(VoxelSpaceFile.class.getName()).log(Level.SEVERE, null, ex);
            }
        } else {
            System.err.println("The voxel space file is null.");
            return null;
        }
        
        return parsed;
    }
    
}
